﻿Option Strict On
'**********************************************************************************************
'* AdvancedHMI Driver
'* http://www.advancedhmi.com
'* ModbusTCP driver - Winforms
'*
'* Archie Jacobs
'* Manufacturing Automation, LLC
'* support@advancedhmi.com
'* 13-FEB-16
'*
'*
'* Copyright 2016 Archie Jacobs
'*
'* This class creates and interface between an Ethernet/IP driver and the AdvancedHMI
'* visual controls.
'*
'* NOTICE : If you received this code without a complete AdvancedHMI solution
'* please report to sales@advancedhmi.com
'*
'* Distributed under the GNU General Public License (www.gnu.org)
'*
'* This program is free software; you can redistribute it and/or
'* as published by the Free Software Foundation; either version 2
'* of the License, or (at your option) any later version.
'*
'* This program is distributed in the hope that it will be useful,
'* but WITHOUT ANY WARRANTY; without even the implied warranty of
'* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'* GNU General Public License for more details.

'* You should have received a copy of the GNU General Public License
'* along with this program; if not, write to the Free Software
'* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
'*
'* 13-FEB-16 New architecture further separating UI element connection from driver
'*******************************************************************************************************
Imports System.ComponentModel.Design
Imports System.Windows.Forms

'<CLSCompliant(True)>
<System.ComponentModel.DefaultEvent("DataReceived")>
Public Class ModbusTCPCom
    Inherits MfgControl.AdvancedHMI.Drivers.ModbusTCP
    Implements System.ComponentModel.IComponent
    Implements System.ComponentModel.ISupportInitialize
    Implements System.Windows.Forms.IBindableComponent

    Private Shared ReadOnly EventDisposed As New Object()
    Public Event Disposed As EventHandler Implements System.ComponentModel.IComponent.Disposed

    Protected m_synchronizationContext As System.Threading.SynchronizationContext

#Region "Constructor"
    Public Sub New(ByVal container As System.ComponentModel.IContainer)
        Me.New()

        'Required for Windows.Forms Class Composition Designer support
        container.Add(Me)
    End Sub

    Public Sub New()
        MyBase.New()

        m_synchronizationContext = System.Windows.Forms.WindowsFormsSynchronizationContext.Current
    End Sub


    Protected Overrides Sub Dispose(disposing As Boolean) 'Implements System.ComponentModel.IComponent.Dispose
        MyBase.Dispose(disposing)

        If disposing Then
            SyncLock Me
                If Site IsNot Nothing AndAlso Site.Container IsNot Nothing Then
                    Site.Container.Remove(Me)
                End If
                'If events IsNot Nothing Then
                'Dim handler As EventHandler = DirectCast(events(EventDisposed), EventHandler)
                RaiseEvent Disposed(Me, EventArgs.Empty)
                'End If
            End SyncLock
        End If
    End Sub



    ''Component overrides dispose to clean up the component list.
    'Protected Overrides Sub Dispose(ByVal disposing As Boolean)
    '    MyBase.Dispose(disposing)
    'End Sub
#End Region

#Region "Properties"
    Private m_IniFileName As String = ""
    Public Property IniFileName As String
        Get
            Return m_IniFileName
        End Get
        Set(value As String)
            m_IniFileName = value
        End Set
    End Property

    Private m_IniFileSection As String
    Public Property IniFileSection As String
        Get
            Return m_IniFileSection
        End Get
        Set(value As String)
            m_IniFileSection = value
        End Set
    End Property



    Public Overrides Property IPAddress As String
        Get
            Return MyBase.IPAddress
        End Get
        Set(value As String)
            If Not String.IsNullOrEmpty(value) Then
                If MyBase.IPAddress <> value Then
                    MyBase.IPAddress = value
                Else
                    'CreateDLLInstance()
                End If
            Else
                MyBase.IPAddress = ""
            End If

        End Set
    End Property

    Private m_site As System.ComponentModel.ISite
    <System.ComponentModel.Browsable(False), System.ComponentModel.DesignerSerializationVisibility(System.ComponentModel.DesignerSerializationVisibility.Hidden)>
    Public Overridable Property Site() As System.ComponentModel.ISite Implements System.ComponentModel.IComponent.Site
        Get
            Return m_site
        End Get
        Set(value As System.ComponentModel.ISite)
            m_site = value
        End Set
    End Property


    <System.ComponentModel.Browsable(False), System.ComponentModel.DesignerSerializationVisibility(System.ComponentModel.DesignerSerializationVisibility.Hidden)>
    Protected ReadOnly Property DesignMode() As Boolean
        Get
            Dim s As System.ComponentModel.ISite = m_site
            If s Is Nothing Then
                Return False
            Else
                Return s.DesignMode
            End If
            ' Return If((s Is Nothing), False, s.DesignMode)
        End Get
    End Property
#End Region

#Region "Binding Properties"
    Private m_bindingContext As BindingContext
    Private m_dataBindings As ControlBindingsCollection
    <System.ComponentModel.DesignerSerializationVisibility(System.ComponentModel.DesignerSerializationVisibility.Content)>
    Public ReadOnly Property DataBindings As ControlBindingsCollection Implements IBindableComponent.DataBindings
        Get
            If (m_dataBindings Is Nothing) Then
                m_dataBindings = New ControlBindingsCollection(Me)
            End If
            Return m_dataBindings
        End Get
    End Property

    <System.ComponentModel.DesignerSerializationVisibility(System.ComponentModel.DesignerSerializationVisibility.Content)>
    Public Property BindingContext As BindingContext Implements IBindableComponent.BindingContext
        Get
            If m_bindingContext Is Nothing Then
                m_bindingContext = New BindingContext()
            End If
            Return m_bindingContext
        End Get
        Set(value As BindingContext)
            m_bindingContext = value
        End Set
    End Property
#End Region

    Protected Overridable Function GetService(service As Type) As Object
        Dim s As System.ComponentModel.ISite = Site
        Return (If((s Is Nothing), Nothing, s.GetService(service)))
    End Function


#Region "Events"
    Protected Overrides Sub OnConnectionClosed(ByVal e As EventArgs)
        If m_synchronizationContext IsNot Nothing Then  ' m_SynchronizingObject IsNot Nothing Then
            Try
                'm_SynchronizingObject.BeginInvoke(occd, Parameters)
                m_synchronizationContext.Post(AddressOf ConnectionClosedSync, EventArgs.Empty)
            Catch ex As Exception
            End Try
            'End If
        Else
            MyBase.OnConnectionClosed(e)
        End If
    End Sub

    Private Sub ConnectionClosedSync(ByVal e As Object)
        Dim e1 As EventArgs = DirectCast(e, EventArgs)
        MyBase.OnConnectionClosed(e1)
    End Sub
    '********************************************************************************************************************************


    Protected Overrides Sub OnConnectionEstablished(ByVal e As EventArgs)
        If m_synchronizationContext IsNot Nothing Then
            'If TypeOf (m_SynchronizingObject) Is System.Windows.Forms.Control AndAlso DirectCast(m_SynchronizingObject, System.Windows.Forms.Control).IsHandleCreated Then
            'm_SynchronizingObject.BeginInvoke(oced, Parameters)
            m_synchronizationContext.Post(AddressOf ConnectionEstabishedSync, EventArgs.Empty)
            'End If
            'End If
        Else
            MyBase.OnConnectionEstablished(e)
        End If
    End Sub


    Private Sub ConnectionEstabishedSync(ByVal e As Object)
        Dim e1 As EventArgs = DirectCast(e, EventArgs)
        MyBase.OnConnectionEstablished(e1)
    End Sub


    Protected Overrides Sub OnDataReceived(ByVal e As MfgControl.AdvancedHMI.Drivers.Common.PlcComEventArgs)
        If m_synchronizationContext IsNot Nothing Then
            'm_SynchronizingObject.BeginInvoke(drsd, New Object() {Me, e})
            m_synchronizationContext.Post(AddressOf DataReceivedSync, e)
        Else
            MyBase.OnDataReceived(e)
        End If
    End Sub


    Private Sub DataReceivedSync(ByVal e As Object)
        Try
            Dim e1 As MfgControl.AdvancedHMI.Drivers.Common.PlcComEventArgs = DirectCast(e, MfgControl.AdvancedHMI.Drivers.Common.PlcComEventArgs)
            MyBase.OnDataReceived(e1)
        Catch ex As Exception
            Dim dbg = 0
        End Try
    End Sub


    '***********************************************************************************************************
    Protected Overrides Sub OnComError(ByVal e As MfgControl.AdvancedHMI.Drivers.Common.PlcComEventArgs)
        If m_synchronizationContext IsNot Nothing Then
            Dim Parameters() As Object = {Me, e}
            'm_SynchronizingObject.BeginInvoke(errorsd, Parameters)
            m_synchronizationContext.Post(AddressOf ErrorReceivedSync, e)
        Else
            MyBase.OnComError(e)
        End If
    End Sub

    Private Sub ErrorReceivedSync(ByVal e As Object)
        Dim e1 As MfgControl.AdvancedHMI.Drivers.Common.PlcComEventArgs = DirectCast(e, MfgControl.AdvancedHMI.Drivers.Common.PlcComEventArgs)
        MyBase.OnComError(e1)
    End Sub
    '***********************************************************************************************************
    '***********************************************************************************************************
    Protected Overrides Sub OnSubscriptionDataReceived(e As MfgControl.AdvancedHMI.Drivers.Common.SubscriptionEventArgs)
        If m_synchronizationContext IsNot Nothing Then
            Dim Parameters() As Object = {Me, e}
            'm_SynchronizingObject.BeginInvoke(e.dlgCallBack, Parameters)
            m_synchronizationContext.Post(AddressOf DataRecSync, e)
        Else
            MyBase.OnSubscriptionDataReceived(e)
        End If
    End Sub

    Private Sub DataRecSync(ByVal e As Object)
        Dim e1 As MfgControl.AdvancedHMI.Drivers.Common.SubscriptionEventArgs = DirectCast(e, MfgControl.AdvancedHMI.Drivers.Common.SubscriptionEventArgs)
        e1.dlgCallBack(Me, e1)
    End Sub

#End Region

#Region "Initializing"
    Private Initializing As Boolean
    Public Sub BeginInit() Implements System.ComponentModel.ISupportInitialize.BeginInit
        Initializing = True
    End Sub

    Public Sub EndInit() Implements System.ComponentModel.ISupportInitialize.EndInit
        If Not Me.DesignMode Then
            If Not String.IsNullOrEmpty(m_IniFileName) Then
                Try
                    Utilities.SetPropertiesByIniFile(Me, m_IniFileName, m_IniFileSection)
                Catch ex As Exception
                    System.Windows.Forms.MessageBox.Show("INI File - " & ex.Message)
                End Try
            End If
        End If
        Initializing = False
    End Sub
#End Region

#Region "Tag Alias"
    Private TagAliasFileRead As Boolean
    Private TagAlias As Dictionary(Of String, String)

    Public Overrides Function BeginRead(ByVal startAddress As String, ByVal numberOfElements As Integer) As Integer
        Return MyBase.BeginRead(ReadTagAliases(startAddress), numberOfElements)
    End Function

    Public Overrides Function BeginWrite(startAddress As String, numberOfElements As Integer, dataToWrite() As String) As Integer
        Return MyBase.BeginWrite(ReadTagAliases(startAddress), numberOfElements, dataToWrite)
    End Function

    Public Overrides Function Subscribe(ByVal plcAddress As String, ByVal numberOfElements As Int16, ByVal pollRate As Integer, ByVal callback As EventHandler(Of MfgControl.AdvancedHMI.Drivers.Common.PlcComEventArgs)) As Integer
        Return MyBase.Subscribe(ReadTagAliases(plcAddress), numberOfElements, pollRate, callback)
    End Function


    Private m_TagAliasFile As String = ".\TagAlias.txt"
    Public Property TagAliasFile As String
        Get
            Return m_TagAliasFile
        End Get
        Set(value As String)

            m_TagAliasFile = value
        End Set
    End Property
    '********************************************************
    '* Check to see if a tag alias "database" exists
    '* If so, read it into a dictionary
    '********************************************************
    Private Function ReadTagAliases(ByVal address As String) As String
        If Not TagAliasFileRead Then
            If Not String.IsNullOrEmpty(m_TagAliasFile) AndAlso (System.IO.File.Exists(m_TagAliasFile)) Then
                TagAlias = New Dictionary(Of String, String)(StringComparer.InvariantCultureIgnoreCase)
                Using sr As New System.IO.StreamReader(m_TagAliasFile)
                    Dim Line As String
                    Dim Items() As String
                    While Not sr.EndOfStream
                        Line = sr.ReadLine()
                        Items = Line.Split(","c)
                        If Items.Length > 1 Then
                            '* Item(0)=Alias, Item(1)=PLC Address
                            TagAlias.Add(Items(0), Items(1))
                        End If
                    End While
                End Using
            End If
            TagAliasFileRead = True
        End If

        If TagAlias IsNot Nothing Then
            If TagAlias.ContainsKey(address) Then
                Return TagAlias(address)
            End If
        End If

        Return address
    End Function

#End Region

End Class
